@@ -1,7 +1,7 @@
 /*
  * dowall.c	- Write to all users on the system.
  *
- * Author:	  Miquel van Smoorenburg, miquels@drinkel.ow.org
+ * Author:	  Miquel van Smoorenburg, miquels@drinkel.cistron.nl
  * 
  * Version:	  1.01  18-11-1992
  *		  - initial version.
@@ -14,8 +14,11 @@
  *		  - Added some more code to prevent 'hanging' on
  *		    open()'s or write()'s.
  *
+ *		  1.3   06-Jul-1996
+ *		  - Fixed "you don't exist" message for uid == 0 (root).
+ *
  *		This file is part of the sysvinit suite,
- *		Copyright 1991-1995 Miquel van Smoorenburg.
+ *		Copyright 1991-1996 Miquel van Smoorenburg.
  *
  *		This program is free software; you can redistribute it and/or
  *		modify it under the terms of the GNU General Public License
@@ -64,15 +67,18 @@ int fromshutdown;
   char *tty, *p;
   time_t t;
   struct utmp utmp;
-  
+  int uid;
+
   if ((fp = fopen(UTMP_FILE, "r")) == (FILE *)0) return;
   
   if (init == 0) {
-  	if ((pwd = getpwuid(getuid())) == (struct passwd *)0) {
+	uid = getuid();
+  	if ((pwd = getpwuid(uid)) == (struct passwd *)0 &&
+	    (uid != 0 || fromshutdown)) {
   		fprintf(stderr, "You don't exist. Go away.\n");
   		exit(1);
   	}
-  	user = pwd->pw_name;
+  	user = pwd ? pwd->pw_name : "root";
   	if ((p = ttyname(0)) != (char *)0) {
   		if ((tty = strrchr(p, '/')) != NULL)
   			tty++;
@@ -109,7 +115,7 @@ int fromshutdown;
   			fputs(text, tp);
 #ifdef AEROSMITH
 			if (fromshutdown && !strcmp(utmp.ut_user, "tyler"))
-			 fputs("(Well hello Mr. Tyler - going DOWN?)\r\n", tp);
+			 fputs("(Oh hello Mr. Tyler - going DOWN?)\r\n", tp);
 #endif
   			fclose(tp);
   		} else
