@@ -17,9 +17,9 @@
  *		Reboot and halt are both this program. Reboot
  *		is just a link to halt.
  *
- * Author:	Miquel van Smoorenburg, miquels@drinkel.ow.org
+ * Author:	Miquel van Smoorenburg, miquels@drinkel.cistron.nl
  *
- * Version:	2.1,  16-Feb-1996
+ * Version:	2.3,  13-Aug-1996
  *
  *		This file is part of the sysvinit suite,
  *		Copyright 1991-1996 Miquel van Smoorenburg.
@@ -43,8 +43,9 @@
 #include <signal.h>
 #include <stdio.h>
 #include <getopt.h>
+#include "reboot.h"
 
-char *Version = "@(#)halt 2.1 16-Feb-1996 MvS";
+char *Version = "@(#)halt 2.3 13-Aug-1996 MvS";
 char *progname;
 
 #define COMPATIBLE 1	  /* For compatibility with sysvinit 2.4 */
@@ -140,7 +141,6 @@ char **argv;
 {
   struct utmp wtmp;
   int fd;
-  time_t t;
   int do_reboot = 0;
   int do_sync = 1;
   int do_wtmp = 1;
@@ -202,13 +202,16 @@ char **argv;
 
   /* Record the fact that we're going down */
   if (do_wtmp && (fd = open(WTMP_FILE, O_WRONLY|O_APPEND)) >= 0) {
-  	time(&t);
   	strcpy(wtmp.ut_user, "shutdown");
   	strcpy(wtmp.ut_line, "~");
   	strcpy(wtmp.ut_id,  "~~");
   	wtmp.ut_pid = 0;
   	wtmp.ut_type = RUN_LVL;
-  	wtmp.ut_time = t;
+#if defined(__GNU_LIBRARY__) && __GNU_LIBRARY__ >= 6
+	gettimeofday(&wtmp.ut_tv, NULL);
+#else
+	time(&wtmp.ut_time);     
+#endif                                                
   	write(fd, (char *)&wtmp, sizeof(wtmp));
   	close(fd);
   }
@@ -222,19 +225,19 @@ char **argv;
   }
 
   if (do_reboot) {
-	reboot(0xfee1dead, 672274793, 0x01234567);
+	init_reboot(BMAGIC_REBOOT);
   } else {
 	/* Turn on hard reboot, CTRL-ALT-DEL will reboot now */
-	reboot(0xfee1dead, 672274793, 0x89ABCDEF);
+	init_reboot(BMAGIC_HARD);
 
 	/* Stop init; it is insensitive to the signals sent by the kernel. */
 	kill(1, SIGTSTP);
 
 	/* And perform the `halt' system call. */
-	reboot(0xfee1dead, 672274793, 0xCDEF0123);
+	init_reboot(BMAGIC_HALT);
   }
   /* If we return, we (c)ontinued from the kernel monitor. */
-  reboot(0xfee1dead, 672274793, 0);
+  init_reboot(BMAGIC_SOFT);
   kill(1, SIGCONT);
 
   exit(0);
