@@ -5,7 +5,7 @@
  *		     init [0123456SsQqAaBbCc]
  *		  telinit [0123456SsQqAaBbCc]
  *
- * Version:	@(#)init.c  2.62  29-May-1996  MvS
+ * Version:	@(#)init.c  2.69  30-Nov-1996  MvS
  *
  *		This file is part of the sysvinit suite,
  *		Copyright 1991-1996 Miquel van Smoorenburg.
@@ -24,6 +24,7 @@
 
 /* Debug and test modes */
 #define DEBUG	   0			/* Debug code off */
+#define INITDEBUG  0			/* Fork at startup so we can debug init. */
 #define CLEAN_UTMP 1			/* Do garbage collect in utmp*/
 #undef  ROOTFS	   "/dev/hda2"		/* Root fs to use */
 
@@ -37,7 +38,7 @@
 #define SLEEPTIME  300			/* Disable time */
 
 /* Default path inherited by every child if it's not set. */
-#define PATH_DFL   "PATH=/bin:/sbin:/usr/bin:/usr/sbin"
+#define PATH_DFL   "PATH=/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin"
 
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -64,6 +65,7 @@
 #  include "initreq.h"
 #endif
 #include "paths.h"
+#include "reboot.h"
 
 #ifndef NO_PROCESS
 #  define NO_PROCESS 0
@@ -78,9 +80,9 @@
 				sigaction(sig, &sa, NULL);
 
 /* Version information */
-char *Version = "@(#) init 2.62 29-May-1996 MvS";
-char *bootmsg = "version 2.62 booting";
-#define E_VERSION "INIT_VERSION=sysvinit-2.62"
+char *Version = "@(#) init 2.69 30-Nov-1996 MvS";
+char *bootmsg = "version 2.69 booting";
+#define E_VERSION "INIT_VERSION=sysvinit-2.69"
 
 /* Information about a process in the in-core inittab */
 typedef struct _child_ {
@@ -122,6 +124,8 @@ int sltime = 5;			/* Sleep time between TERM and KILL */
 char *argv0;			/* First arguments; show up in ps listing */
 int maxproclen;			/* Maximal length of argv[0] */
 struct utmp utproto;		/* Only used for sizeof(utproto.ut_id) */
+char *console_dev = CONSOLE;	/* Console device. */
+int wroteiosave = 0;		/* Did we write ioctl.save yet? */
 
 /* Prototypes. */
 void Wtmp(char *user, char *id, int pid, int type, char *line);
@@ -248,7 +252,7 @@ void chld_handler()
   while((pid = waitpid(-1, &st, WNOHANG)) != 0) {
 	if (errno == ECHILD) break;
 #ifdef __alpha__
-	/* Buggy library */
+	/* Some (old) Linux/Alpha version have a buggy library */
 	if (pid == -ECHILD) break;
 #endif
 	for( ch = family; ch; ch = ch->next )
@@ -284,6 +288,15 @@ void cont_handler()
   got_cont = 1;
 }
 
+/*
+ * OOPS: segmentation violation!
+ */
+void segv_handler()
+{
+  Log(L_VB, "PANIC: segmentation violation! giving up..");
+  while(1) pause();
+}
+
 /*
  * The SIGSTOP & SIGTSTP handler
  */
@@ -301,15 +314,15 @@ void SetTerm(int how)
   int fd, fd2;
   int restore_ok = 0;
 
-  if ((fd = serial_open(CONSOLE, O_RDWR|O_NOCTTY)) < 0) {
-	Log(L_VB, "can't open %s", CONSOLE);
+  if ((fd = serial_open(console_dev, O_RDWR|O_NOCTTY)) < 0) {
+	Log(L_VB, "can't open %s", console_dev);
 	return;
   }
 
   /* Do we want to save or restore modes. */
   if (how == 0) {
 	/* Restore. */
-	if ((fd2 = open(IOSAVE, O_RDONLY)) >= 0) {
+	if (wroteiosave && (fd2 = open(IOSAVE, O_RDONLY)) >= 0) {
 		if (read(fd2, &tty, sizeof(tty)) == sizeof(tty))
 			restore_ok = 1;
 		close(fd2);
@@ -319,8 +332,8 @@ void SetTerm(int how)
 		(void) ioctl(fd, TCGETS, &tty);
 
 		/* Set speed etc. */
-		tty.c_cflag = HUPCL|CS8|B9600|CLOCAL;
-		/* FIXME: should we set C_LINE here? */
+		tty.c_cflag &= CBAUD|CBAUDEX|CSIZE|CSTOPB|PARENB|PARODD;
+		tty.c_cflag |= HUPCL|CLOCAL;
 
 		/* Set the most important characters */
 		tty.c_cc[VINTR]  = 3;
@@ -356,6 +369,7 @@ void SetTerm(int how)
 	return;
   }
   (void) write(fd2, &tty, sizeof(struct termios));
+  wroteiosave = 1;
   (void) close(fd);
   (void) close(fd2);
 }
@@ -365,7 +379,7 @@ void print(char *s)
 {
   int fd;
 
-  if ((fd = serial_open(CONSOLE, O_WRONLY|O_NOCTTY|O_NDELAY)) >= 0) {
+  if ((fd = serial_open(console_dev, O_WRONLY|O_NOCTTY|O_NDELAY)) >= 0) {
 	write(fd, s, strlen(s));
 	close(fd);
   }
@@ -430,19 +444,24 @@ int any(char *s1, char *s2)
 
 
 /*
- * Fork and execute.
+ *	Fork and execute.
+ *
+ *	This function is too long and indents too deep.
+ *
  */
 int Spawn(CHILD *ch)
 {
   char *args[16];		/* Argv array */
   char buf[136];		/* Line buffer */
-  int f, pid;			/* Scratch variables */
+  int f, st;			/* Scratch variables */
   char *ptr;			/* Ditto */
   time_t t;			/* System time */
   int oldAlarm;			/* Previous alarm value */
   char *proc = ch->process;	/* Command line */
   char i_lvl[] = "RUNLEVEL=x";	/* Runlevel in environment. */
   char i_prev[] = "PREVLEVEL=x";/* Previous runlevel. */
+  char i_cons[32];		/* console device. */
+  pid_t pid, pgrp;		/* child, console process group. */
 
   /* Skip '+' if it's there */
   if (proc[0] == '+') proc++;
@@ -532,32 +551,77 @@ int Spawn(CHILD *ch)
   args[0] = args[1];
   while(1) {
 	if ((pid = fork()) == 0) {
+		close(0);
+		close(1);
+		close(2);
 		close(PIPE_FD);
-#if DEBUG && !SYSLOG
-		pid = open("/dev/null", O_RDONLY);
-		close(pid);
-		if (pid != 0)
-			Log(L_VB, "some fd is still open (%d)", pid);
-#endif
 		/* Now set RUNLEVEL and PREVLEVEL */
+		sprintf(i_cons, "CONSOLE=%s", console_dev);
 		i_lvl[9]   = thislevel;
 		i_prev[10] = prevlevel;
 		putenv(i_lvl);
 		putenv(i_prev);
+		putenv(i_cons);
 		putenv(E_VERSION);
 
-  		/* The single user entry needs to talk to the console */
-  		if (strcmp(ch->rlevel, "S") == 0) {
+		/*
+		 *	In sysinit, boot, bootwait or single user mode:
+		 *	any wait-type subprocess gets the console
+		 *	as its controlling tty (if not in use yet).
+		 */
+  		if (strchr("*#sS", runlevel) && ch->flags & WAITING) {
+			/*
+			 *	We fork once extra. This is so that we can
+			 *	wait and change the process group and session
+			 *	of the console after exit of the leader.
+			 */
 			setsid();
-			f = serial_open(CONSOLE, O_RDWR);
-			ioctl(f, TIOCSCTTY, (void *)1);
+			f = serial_open(console_dev, O_RDWR|O_NOCTTY);
+			(void)ioctl(f, TIOCSCTTY, 0);
   			dup(f);
   			dup(f);
+			if ((pid = fork()) < 0) {
+  				Log(L_VB, "cannot fork");
+				exit(1);
+			}
+			if (pid > 0) {
+				/*
+				 *	Ignore keyboard signals etc.
+				 */
+				signal(SIGINT, SIG_IGN);
+				signal(SIGQUIT, SIG_IGN);
+				signal(SIGTSTP, SIG_IGN);
+				signal(SIGCHLD, SIG_DFL);
+
+				while(waitpid(pid, &st, 0) != pid)
+					;
+				pgrp = 0;
+				(void)ioctl(f, TIOCGPGRP, &pgrp);
+				if (pgrp != getpid())
+					exit(0);
+				/*
+				 *	De-attach session from controlling tty.
+				 *	Or actually, the other way around.
+				 *	This _without_ SIGHUP etc.
+				 */
+				if ((pid = fork()) < 0) {
+  					Log(L_VB, "cannot fork");
+					exit(1);
+				}
+				if (pid == 0) {
+					setsid();
+					(void)ioctl(f, TIOCSCTTY, 1);
+					exit(0);
+				}
+				while(waitpid(pid, &st, 0) != pid)
+					;
+				exit(0);
+			}
 			/* Set ioctl settings to default ones */
 			SetTerm(0);
   		} else {
-			if ((f = serial_open(CONSOLE, O_RDWR|O_NOCTTY)) < 0) {
-				Log(L_VB, "open(%s): %s", CONSOLE,
+			if ((f = serial_open(console_dev, O_RDWR|O_NOCTTY)) < 0) {
+				Log(L_VB, "open(%s): %s", console_dev,
 					sys_errlist[errno]);
 				f = open("/dev/null", O_RDWR);
 			}
@@ -661,7 +725,9 @@ void ReadItab(void)
   int talk;			/* Talk to the user */
   int done = 0;			/* Ready yet? */
   sigset_t omask;		/* For blocking SIGCHLD. */
+#ifdef INITLVL
   struct stat st;		/* To stat INITLVL */
+#endif
 
 #if DEBUG
   if (newFamily != NULL) {
@@ -705,11 +771,11 @@ void ReadItab(void)
 	if (!id || !*id) strcpy(err, "missing id field");
 	if (!rlevel)     strcpy(err, "missing runlevel field");
 	if (!process)    strcpy(err, "missing process field");
-	if (!action && !*action)
+	if (!action || !*action)
 			strcpy(err, "missing action field");
 	if (id && strlen(id) > sizeof(utproto.ut_id))
 			sprintf(err, "id field too long (max %d characters)",
-				sizeof(utproto.ut_id));
+				(int)sizeof(utproto.ut_id));
 	if (rlevel && strlen(rlevel) > 11)
 			strcpy(err, "rlevel field too long (max 11 characters)");
 	if (process && strlen(process) > 127) strcpy(err, "process field too long");
@@ -901,8 +967,9 @@ void ReadItab(void)
     }
   }
 
-  /* Now give all processes the chance to die and collect exit statuses */
-  /* FIXME: how to give away time slice?? */
+  /*
+   *	Now give all processes the chance to die and collect exit statuses.
+   */
   if (foundOne) Sleep(1);
   for(ch = family; ch; ch = ch->next)
 	if (ch->flags & KILLME) {
@@ -920,8 +987,7 @@ void ReadItab(void)
 	}
 
   /* Both rounds done; clean up the list. */
-  omask = siggetmask();
-  sigblock(sigmask(SIGCHLD));
+  omask = sigblock(sigmask(SIGCHLD));
   for(ch = family; ch; ch = old) {
 	old = ch->next;
 	free(ch);
@@ -931,6 +997,7 @@ void ReadItab(void)
   newFamily = NULL;
   sigsetmask(omask);
 
+#ifdef INITLVL
   /* Dispose of INITLVL file. */
   if (lstat(INITLVL, &st) >= 0 && S_ISLNK(st.st_mode)) {
 	/* INITLVL is a symbolic link, so just truncate the file. */
@@ -939,6 +1006,17 @@ void ReadItab(void)
 	/* Delete INITLVL file. */
   	unlink(INITLVL);
   }
+#endif
+#ifdef INITLVL2
+  /* Dispose of INITLVL2 file. */
+  if (lstat(INITLVL2, &st) >= 0 && S_ISLNK(st.st_mode)) {
+	/* INITLVL2 is a symbolic link, so just truncate the file. */
+	close(open(INITLVL2, O_WRONLY|O_TRUNC));
+  } else {
+	/* Delete INITLVL2 file. */
+  	unlink(INITLVL2);
+  }
+#endif
 }
 
 /*
@@ -1001,8 +1079,11 @@ int AskRunLevel()
 
 
   SetTerm(0);
-  fd = serial_open(CONSOLE, O_RDWR|O_NOCTTY);
-	/* FIXME as soon as O_NDELAY != O_NONBLOCK */
+  /*
+   *	FIXME: should we open the tty O_NONBLOCK and CLOCAL mode?
+   */
+  fd = serial_open(console_dev, O_RDWR|O_NOCTTY);
+
   if (fd < 0) return('S');
 
   while(!strchr("0123456789S", lvl)) {
@@ -1053,7 +1134,11 @@ int GetInitDefault(void)
 
 
 /*
- * We got signaled: read the new level from INITLVL
+ *	We got signaled.
+ *
+ *	Do actions for the new level. If we are compatible with
+ *	the "old" INITLVL and arg == 0, try to read the new
+ *	runlevel from that file first.
  */
 int ReadLevel(int arg)
 {
@@ -1061,17 +1146,26 @@ int ReadLevel(int arg)
   int st;				/* Sleep time */
   CHILD *ch;				/* Walk through list */
   FILE *fp;				/* File pointer for INITLVL */
-  int ok = 0;
+  int ok = 1;
   struct stat stt;
 
 #if CLEAN_UTMP
   /* Clean up utmp file first. */
-  Wtmp(NULL, NULL, -2, 0, NULL);
+  Wtmp(NULL, "", -2, 0, NULL);
 #endif
 
+#ifdef INITLVL
+  ok = 0;
+
   if (arg == 0) {
-	if ((fp = fopen(INITLVL, "r")) == NULL ||
-	      (stat(INITLVL, &stt) == 0 && stt.st_size == 0L)) {
+	fp = NULL;
+	if (stat(INITLVL, &stt) != 0 || stt.st_size != 0L)
+		fp = fopen(INITLVL, "r");
+#ifdef INITLVL2
+	if (fp == NULL && (stat(INITLVL2, &stt) != 0 || stt.st_size != 0L))
+		fp = fopen(INITLVL2, "r");
+#endif
+	if (fp == NULL) {
 		/* INITLVL file is empty or not there - act as 'init q' */
 		Log(L_SY, "Re-reading inittab");
   		return(runlevel);
@@ -1084,12 +1178,14 @@ int ReadLevel(int arg)
 	ok = 1;
   }
   if (islower(foo)) foo = toupper(foo);
+  if (ok == 2) sltime = st;
+
+#endif /* INITLVL */
 
   if (ok < 1 || ok > 2 || strchr("QS0123456789ABC", foo) == NULL) {
  	Log(L_VB, "bad runlevel: %c", foo);
   	return(runlevel);
   }
-  if (ok == 2) sltime = st;
 
   /* Be verbose 'bout it all */
   switch(foo) {
@@ -1139,6 +1235,17 @@ int ReadLevel(int arg)
 }
 
 
+/*
+ * Zero out an utmp struct but copy the id field.
+ */
+void EmptyStruct(struct utmp *oldu, struct utmp *newu)
+{
+  memset(newu, 0, sizeof(struct utmp));
+  strncpy(newu->ut_id, oldu->ut_id, sizeof(newu->ut_id));
+  strncpy(newu->ut_line, oldu->ut_line, sizeof(newu->ut_line));
+  newu->ut_type = DEAD_PROCESS;
+}
+
 /*
  * Log an event ONLY in the wtmp file (reboot, runlevel)
  */
@@ -1151,7 +1258,6 @@ char *line)			/* Which line is this */
 {
   int fd;
   struct utmp utmp;
-  time_t t;
 
   if ((fd = open(WTMP_FILE, O_WRONLY|O_APPEND)) < 0) return;
 
@@ -1171,8 +1277,11 @@ char *line)			/* Which line is this */
   memset(&utmp, 0, sizeof(utmp));
 
   /* Enter new fields */
-  time(&t);
-  utmp.ut_time = t;
+#if defined(__GNU_LIBRARY__) && __GNU_LIBRARY__ >= 6
+  gettimeofday(&utmp.ut_tv, NULL);
+#else
+  time(&utmp.ut_time);
+#endif
   utmp.ut_pid  = pid;
   utmp.ut_type = type;
   strncpy(utmp.ut_name, user, sizeof(utmp.ut_name));
@@ -1185,6 +1294,18 @@ char *line)			/* Which line is this */
 
 /*
  * Log an event into the WTMP and UTMP files.
+ *
+ * FIXME: we shouldn't zero out other processes' utmp entries. What we
+ *        should do to get/update an utmp entry:
+ *        o see if an entry with matching ut_id field is present
+ *        o if so, re-use it.
+ *        o if not, re-open utmp in O_APPEND mode and add new entry.
+ *
+ * Thanks to Michael Haardt for pointing this out.
+ *
+ * Also, check if the RUNLEVEL entry is still present and rewrite if
+ * it is missing. (People tend to mess up their utmp files).
+ *
  */
 void Wtmp(
 char *user,			/* name of user */
@@ -1200,17 +1321,13 @@ char *line)			/* LINE if used. */
   int fd2;			/* File Descriptor for WTMP */
   int found = 0;		/* Was the record found in UTMP */
   int freeEntry = -1;		/* Was a free entry found during UTMP scan? */
+  int matchEntry = -1;		/* Was a matching entry found ? */
   int lineno;			/* Offset into UTMP file */
-  time_t t;			/* What's the time? */
   static int opened = 0;	/* Already opened /var/run/utmp? */
 #if CLEAN_UTMP
   CHILD *ch;			/* Scratch */
 #endif
 
-  /* Zero the empty structure */
-  memset(&empty, 0, sizeof(struct utmp));
-  empty.ut_type = NO_PROCESS;
-
   /* Clear utmp if not yet opened. */
   if (!opened) close(open(UTMP_FILE, O_WRONLY|O_CREAT|O_TRUNC, 0644));
 
@@ -1231,6 +1348,7 @@ char *line)			/* LINE if used. */
 			if (type == DEAD_PROCESS) {
 				/* Zero all entries with this pid */
 				(void) lseek(fd, (long) lineno, SEEK_SET);
+				EmptyStruct(&tmp, &empty);
 				write(fd, (char *)&empty, sizeof(struct utmp));
 			} else {
 				(void) lseek(fd, (long) lineno, SEEK_SET);
@@ -1253,14 +1371,25 @@ char *line)			/* LINE if used. */
 #endif
 				/* Zero the entry with this pid */
 				(void) lseek(fd, (long) lineno, SEEK_SET);
-				write(fd, (char *)&empty, sizeof(struct utmp));
+				EmptyStruct(&tmp, &empty);
+				if (write(fd, (char *)&empty, sizeof(struct utmp))
+					!= sizeof(struct utmp)) {
+						Log(L_VB, "error writing utmp struct");
+						break;
+				}
 			}
 		}
 #endif
-		/* See if this is a free entry, save it for later */
-		if ((tmp.ut_pid == 0 && tmp.ut_type != RUN_LVL) ||
-		    tmp.ut_type == 0)
+		/* See if this is a unused entry, save it for later */
+		if ((tmp.ut_pid == 0 && tmp.ut_id[0] == 0
+			&& tmp.ut_user[0] == 0 && tmp.ut_type != RUN_LVL) || tmp.ut_type == 0)
 			if (freeEntry < 0) freeEntry = lineno;
+
+		/* See if this entry is dead and matches our "id", save it for later. */
+		if (tmp.ut_type == DEAD_PROCESS && strncmp(tmp.ut_id, id, sizeof(tmp.ut_id)) == 0
+			&& tmp.ut_user[0] == 0)
+			if (matchEntry < 0) matchEntry = lineno;
+
 		lineno += sizeof(tmp);
 	}
 #if CLEAN_UTMP
@@ -1272,6 +1401,10 @@ char *line)			/* LINE if used. */
 #endif
   }
 
+#if DEBUG
+  Log(L_VB, "Wtmp(): found=%d, match=%d, free=%d", found, matchEntry, freeEntry);
+#endif
+
   if (!found) { /* Enter some defaults */
 	/* Zero the fields */
 	memset(&utmp, 0, sizeof(utmp));
@@ -1282,14 +1415,18 @@ char *line)			/* LINE if used. */
 	strcpy (utmp.ut_line, "");
 
 	/* Where to write new utmp record */
+	if (matchEntry >= 0) freeEntry = matchEntry;
 	if (freeEntry >= 0)
 		lseek(fd, (long) freeEntry, SEEK_SET);
   }
 
   /* Change the values of some fields */
-  time(&t);
+#if defined(__GNU_LIBRARY__) && __GNU_LIBRARY__ >= 6
+  gettimeofday(&utmp.ut_tv, NULL);
+#else
+  time(&utmp.ut_time);
+#endif
   utmp.ut_type = type;
-  utmp.ut_time = t;
   utmp.ut_pid = pid;
   if (line) strncpy(utmp.ut_line, line, sizeof(utmp.ut_line));
 
@@ -1413,6 +1550,7 @@ void FifoNewLevel(int level)
 	oldlevel = runlevel;
 	runlevel = ReadLevel(level);
 	if (oldlevel != 'S' && runlevel == 'S') SetTerm(0);
+	if (runlevel == '6' || runlevel == '0' || runlevel == '1') SetTerm(0);
 	ReadItab();
 	FailCancel();
 	setproctitle("init [%c]", runlevel);
@@ -1423,7 +1561,8 @@ void FifoNewLevel(int level)
  * Read from the init FIFO. Processes like telnetd and rlogind can
  * ask us to create login processes on their behalf.
  *
- * FIXME: this needs to be finished.
+ * FIXME: this needs to be finished. NOT that it is buggy, but we need
+ *	  to add the telnetd/rlogind stuff so people can start using it.
  */
 void CheckInitFifo(void)
 {
@@ -1537,6 +1676,9 @@ void BootTransitions()
 #if DEBUG
 			Log(L_VB, "SYSINIT -> BOOT");
 #endif
+			/* Save tty modes. */
+			SetTerm(1);
+
 			/* Write a boot record. */
 			wroteReboot = 0;
 			WtmpOnly("reboot", "~~", 0, BOOT_TIME, "~");
@@ -1554,6 +1696,9 @@ void BootTransitions()
 #if DEBUG
 			Log(L_VB, "BOOT -> NORMAL");
 #endif
+			/* Save tty modes. */
+			SetTerm(1);
+
 			if (runlevel != newlevel)
 				loglevel = newlevel;
 			runlevel = newlevel;
@@ -1566,6 +1711,7 @@ void BootTransitions()
 #endif
 			/* Save tty modes. */
 			SetTerm(1);
+
 			newlevel = GetInitDefault();
 			if (!did_boot && newlevel != 'S')
 				runlevel = '*';
@@ -1582,11 +1728,13 @@ void BootTransitions()
 		default:
 			loglevel = -1;
 			Log(L_VB, "no more processes left in this runlevel");
+			if (got_signals == 0) {
 #if NEWINIT
-			CheckInitFifo();
+				CheckInitFifo();
 #else
-			pause();
+				pause();
 #endif
+			}
 			break;
 	}
 	if (loglevel > 0) {
@@ -1691,8 +1839,11 @@ void ProcessSignals()
 	{
 		/* We need to go into a new runlevel */
 		oldlevel = runlevel;
+#ifdef INITLVL
 		runlevel = ReadLevel(0);
+#endif
 		if (oldlevel != 'S' && runlevel == 'S') SetTerm(0);
+		if (runlevel == '6' || runlevel == '0' || runlevel == '1') SetTerm(0);
 		ReadItab();
 		FailCancel();
 		setproctitle("init [%c]", runlevel);
@@ -1709,6 +1860,19 @@ int InitMain()
   int f, st;
   CHILD *ch;
   struct sigaction sa;
+  char *s;
+
+#if INITDEBUG
+  /*
+   * Fork so we can debug the init process.
+   */
+  if ((f = fork()) > 0) {
+	while(wait(&st) != f)
+		;
+	write(1, "PRNT: init killed.\r\n", 20);
+	while(1) pause();
+  }
+#endif
 
 #ifdef ROOTFS
   /* Close all file descriptors. */
@@ -1730,7 +1894,7 @@ int InitMain()
 #endif
 
   /* Tell the kernel to send us SIGINT when CTRL-ALT-DEL is pressed */
-  reboot(0xfee1dead, 672274793, 0);
+  init_reboot(BMAGIC_SOFT);
 
   /* Tell the kernel that we want to handle keyboard signals. */
 #ifdef KDSIGACCEPT
@@ -1750,15 +1914,17 @@ int InitMain()
   SETSIG(sa, SIGINT,   signal_handler);
   SETSIG(sa, SIGPWR,   signal_handler);
   SETSIG(sa, SIGWINCH, signal_handler);
-  SETSIG(sa, SIGCHLD,  chld_handler);
   SETSIG(sa, SIGSTOP,  stop_handler);
   SETSIG(sa, SIGTSTP,  stop_handler);
   SETSIG(sa, SIGCONT,  cont_handler);
+  SETSIG(sa, SIGSEGV,  segv_handler);
 
   /* Close whatever files are open, and reset the console. */
   close(0);
   close(1);
   close(2);
+  if ((s = getenv("CONSOLE")) != NULL)
+	console_dev = s;
   SetTerm(0);
   setsid();
 
@@ -1773,11 +1939,13 @@ int InitMain()
 
   /* See if we have to start an emergency shell. */
   if (emerg_shell) {
+	SETSIG(sa, SIGCHLD, SIG_DFL);
 	if ((f = Spawn(&ch_emerg)) > 0) {
 		while(wait(&st) != f)
 			;
 	}
   }
+  SETSIG(sa, SIGCHLD,  chld_handler);
 
   /* Start normal boot procedure. */
   runlevel = '#';
@@ -1863,9 +2031,9 @@ int main(int argc, char *argv[])
   	p = argv[0];
   	
 
-  /* See if I want to become "father of all processes" */
+  /* See if I want to become "parental guide of all processes" */
 #ifndef TEST
-  if (getpid() == INITPID) {
+  if (getpid() == INITPID || strcmp(p, "init.new") == 0 || strcmp(p, "sh") == 0) {
 #else
 			   {
 #endif
@@ -1889,10 +2057,6 @@ int main(int argc, char *argv[])
 	argv0 = argv[0];
 	argv[1] = NULL;
 	setproctitle("init boot");
-#ifdef TEST
-	sleep(5);
-	exit(0);
-#endif
 	InitMain(dflLevel);
   }
 
@@ -1936,8 +2100,14 @@ int main(int argc, char *argv[])
 	alarm(0);
 	return 0;
   }
+#ifndef INITLVL
+  perror(INIT_FIFO);
+  exit(1);
+#endif
   /* Fallthrough to the old method. */
 #endif
+
+#ifdef INITLVL
   /* Now write the new runlevel. */
   if ((fp = fopen(INITLVL, "w")) == NULL) {
 	fprintf(stderr, "%s: cannot create %s\n", p, INITLVL);
@@ -1950,4 +2120,5 @@ int main(int argc, char *argv[])
   if (kill(INITPID, SIGHUP) < 0) perror(p);
 
   exit(0);
+#endif
 }
