@@ -4,9 +4,9 @@
  *		just felt like writing this. No thanks :-).
  *		Also, this version gives lots more info (especially with -x)
  *
- * Author:	Miquel van Smoorenburg, miquels@drinkel.ow.org
+ * Author:	Miquel van Smoorenburg, miquels@drinkel.cistron.nl
  *
- * Version:	@(#)last  2.0  12-Mar-1996  MvS
+ * Version:	@(#)last  2.3  13-Aug-1996  MvS
  *
  *		This file is part of the sysvinit suite,
  *		Copyright 1991-1996 Miquel van Smoorenburg.
@@ -28,10 +28,11 @@
 #include <string.h>
 #include <signal.h>
 #include <getopt.h>
+#include <netinet/in.h>
 #include <netdb.h>
 #include <arpa/inet.h>
 
-char *Version = "@(#) last 2.0 12-Mar-1996 MvS";
+char *Version = "@(#) last 2.3 13-Aug-1996 MvS";
 
 #define CHOP_DOMAIN	0	/* Define to chop off local domainname. */
 
@@ -114,7 +115,7 @@ void quit_handler()
 }
 
 /* Get the basename of a filename */
-char *basename(char *s)
+char *mybasename(char *s)
 {
   char *p;
 
@@ -154,7 +155,7 @@ int list(struct utmp *p, time_t t, int what)
   char length[32];
   char final[128];
   char domain[256];
-  time_t secs;
+  time_t secs, tmp;
   int mins, hours, days;
   char *s, **walk;
 
@@ -173,7 +174,8 @@ int list(struct utmp *p, time_t t, int what)
   }
 
   /* Calculate times */
-  strcpy(logintime, ctime(&p->ut_time));
+  tmp = (time_t)p->ut_time;
+  strcpy(logintime, ctime(&tmp));
   logintime[16] = 0;
   lastdate = p->ut_time;
   sprintf(logouttime, "- %s", ctime(&t) + 11);
@@ -212,7 +214,7 @@ int list(struct utmp *p, time_t t, int what)
 	dns_lookup(domain, p->ut_host, p->ut_addr);
   else {
 	strncpy(domain, p->ut_host, UT_HOSTSIZE);
-	domain[16] = 0;
+	domain[UT_HOSTSIZE] = 0;
   }
 
   if (showhost) {
@@ -271,7 +273,7 @@ struct utmp *get_one(FILE *fp)
   static int pos = -1;
   static int off = -1;
   static int first = 1;
-  int sz;
+  size_t sz;
   static struct utmp buf[UBUFS];
   struct utmp *u;
 
@@ -329,7 +331,7 @@ int main(int argc, char **argv)
   int down;		/* Down flag */
   char *altufile = NULL;/* Alternate wtmp */
 
-  progname = basename(argv[0]);
+  progname = mybasename(argv[0]);
 
   /* Process the arguments. */
   while((c = getopt(argc, argv, "f:n:Rxad0123456789")) != EOF)
@@ -400,8 +402,9 @@ int main(int argc, char **argv)
 
   /* Open the utmp file */
   if ((fp = fopen(ufile, "r")) == NULL) {
+	x = errno;
 	fprintf(stderr, "%s: %s: %s\n", progname, ufile, sys_errlist[errno]);
-	if (altufile == NULL) fprintf(stderr,
+	if (altufile == NULL && x == ENOENT) fprintf(stderr,
 		"Perhaps this file was removed by the operator to prevent logging %s info.\n",
 				progname);
 	exit(1);
@@ -522,7 +525,7 @@ int main(int argc, char **argv)
       if (fseek(fp, -2 * sizeof(struct utmp), SEEK_CUR) < 0) break;
 #endif
   }
-  printf("\n%s begins %s", basename(ufile), ctime(&begintime));
+  printf("\n%s begins %s", mybasename(ufile), ctime(&begintime));
 
   fclose(fp);
   /* Should we free memory here? Nah. */
