@@ -3,7 +3,7 @@
  *
  * Version:	@(#) paths.h 1.71 28-Dec-1995
  *
- * Author:	Miquel van Smoorenburg, <miquels@drinkel.ow.org>
+ * Author:	Miquel van Smoorenburg, <miquels@drinkel.cistron.nl>
  *
  *		This file is part of the sysvinit suite,
  *		Copyright 1991-1995 Miquel van Smoorenburg.
@@ -14,6 +14,7 @@
  *		2 of the License, or (at your option) any later version.
  */
 #define INITLVL		"/etc/initrunlvl"	/* COMPAT: New runlevel for init */
+#define INITLVL2	"/var/log/initrunlvl"	/* COMPAT: New runlevel for init */
 #define PWRSTAT		"/etc/powerstatus"	/* COMPAT: SIGPWR reason (OK/BAD) */
 #define VT_MASTER	"/dev/tty0"		/* Virtual console master */
 #define CONSOLE		"/dev/console"		/* Logical system console */
