@@ -10,9 +10,9 @@
  *		  -c: cancel an already running shutdown.
  *		  -t secs: delay between SIGTERM and SIGKILL for init.
  *
- * Author:	Miquel van Smoorenburg, miquels@drinkel.ow.org
+ * Author:	Miquel van Smoorenburg, miquels@drinkel.cistron.nl
  *
- * Version:	@(#)shutdown  2.2  04-Jun-1996  MvS
+ * Version:	@(#)shutdown  2.4  13-Aug-1996  MvS
  *
  *		This file is part of the sysvinit suite,
  *		Copyright 1991-1996 Miquel van Smoorenburg.
@@ -34,8 +34,9 @@
 #include <fcntl.h>
 #include <utmp.h>
 #include "paths.h"
+#include "reboot.h"
 
-char *Version = "@(#) shutdown 2.2 04-Jun-1996 MvS";
+char *Version = "@(#) shutdown 2.4 13-Aug-1996 MvS";
 
 #define MESSAGELEN	256
 
@@ -126,7 +127,6 @@ void fastdown()
   int do_halt = (down_level[0] == '0');
   char *script;
   char cmd[128];
-  time_t t;
   int fd, i;
   struct utmp wtmp;
 
@@ -180,13 +180,16 @@ void fastdown()
 
   /* Record the fact that we're going down */
   if ((fd = open(WTMP_FILE, O_WRONLY|O_APPEND)) >= 0) {
-        time(&t);
         strcpy(wtmp.ut_user, "shutdown");
         strcpy(wtmp.ut_line, "~");
         strcpy(wtmp.ut_id,  "~~");
         wtmp.ut_pid = 0;
         wtmp.ut_type = RUN_LVL;
-        wtmp.ut_time = t;
+#if defined(__GNU_LIBRARY__) && __GNU_LIBRARY__ >= 6
+        gettimeofday(&wtmp.ut_tv, NULL);
+#else
+        time(&wtmp.ut_time);
+#endif
         write(fd, (char *)&wtmp, sizeof(wtmp));
         close(fd);
   }
@@ -203,11 +206,11 @@ void fastdown()
   /* We're done, halt or reboot now. */
   if (do_halt) {
 	fprintf(stderr, "The system is halted. Press CTRL-ALT-DEL or turn off power\r\n");
-	reboot(0xfee1dead, 672274793, 0x89ABCDEF);
+	init_reboot(BMAGIC_HALT);
 	exit(0);
   }
   fprintf(stderr, "Please stand by while rebooting the system.\r\n");
-  reboot(0xfee1dead, 672274793, 0x01234567);
+  init_reboot(BMAGIC_REBOOT);
   exit(0);
 }
 
